@@ -18,25 +18,24 @@ from pathlib import Path
 from types import SimpleNamespace
 from typing import Optional
 
-import httpx
 from tqdm import tqdm
 
-from .. import exceptions
+from .. import __version__, exceptions
 from . import adapters, fs  # noqa
 
 PREFIXES_TABLE = {
-    'k': 1000,
-    'K': 1000,
-    'Ki': 1024,
-    'ki': 1024,
-    'M': 1000**2,
-    'm': 1000**2,
-    'Mi': 1024**2,
-    'mi': 1024**2,
-    'g': 1000**3,
-    'G': 1000**3,
-    'gi': 1024**3,
-    'Gi': 1024**3,
+    'k': 1_000,
+    'K': 1_000,
+    'Ki': 1_024,
+    'ki': 1_024,
+    'M': 1_000**2,
+    'm': 1_000**2,
+    'Mi': 1_024**2,
+    'mi': 1_024**2,
+    'g': 1_000**3,
+    'G': 1_000**3,
+    'gi': 1_024**3,
+    'Gi': 1_024**3,
 }
 UNITS_TABLE = {'B': 1, 'b': Decimal('0.125')}
 HUMAN_SIZE_REGEX = (
@@ -44,7 +43,6 @@ HUMAN_SIZE_REGEX = (
     % "|".join(PREFIXES_TABLE)
 )
 
-
 logger = logging.getLogger(__name__)
 
 
@@ -78,17 +76,16 @@ def _read_bytes(arg):
 # Still, we'll use os.environb when available.
 def _get_environb(var, default=None):
     try:
-        value = os.environb.get(os.fsencode(var), default)
+        return os.environb.get(os.fsencode(var), default)
     except AttributeError:
-        value = os.environ.get(var, default)
-        if value is not default:
+        if (value := os.environ.get(var, default)) is not default:
             return os.fsencode(value)
-    finally:
-        return value
+        else:
+            return value
 
 
-common_options = argparse.ArgumentParser(add_help=False)
-common_options.add_argument(
+repository_parser = argparse.ArgumentParser(add_help=False)
+repository_parser.add_argument(
     '-r',
     '--repository',
     type=_backend_tuple,
@@ -97,28 +94,37 @@ common_options.add_argument(
     "variable is used as a fallback. If neither is provided, we'll use the CWD.",
     default=os.environ.get('REPLICAT_REPOSITORY', str(Path())),
 )
-common_options.add_argument(
+
+common_options_parser = argparse.ArgumentParser(add_help=False)
+common_options_parser.add_argument(
     '-q',
     '--hide-progress',
     dest='quiet',
     action='store_true',
     help='Disable progress bar for commands that support it.',
 )
-common_options.add_argument(
+common_options_parser.add_argument(
     '-c',
     '--concurrent',
     default=5,
     type=int,
     help='The number of concurrent connections to the backend.',
 )
-common_options.add_argument('-v', '--verbose', action='count', default=0)
-common_options.add_argument(
+cache_options = common_options_parser.add_mutually_exclusive_group()
+cache_options.add_argument(
+    '--cache-directory', help='Cache directory', default=fs.DEFAULT_CACHE_DIRECTORY
+)
+cache_options.add_argument(
+    '--no-cache', action='store_const', const=None, dest='cache_directory'
+)
+common_options_parser.add_argument('-v', '--verbose', action='count', default=0)
+common_options_parser.add_argument(
     '-K', '--key-file', metavar='KEYFILE', dest='key', type=_read_bytes
 )
 # All the different ways to provide the repository password.
 # We could add a proper description for this group, but there's
 # a long-standing argparse bug https://bugs.python.org/issue16807
-password_options = common_options.add_mutually_exclusive_group()
+password_options = common_options_parser.add_mutually_exclusive_group()
 password_options.add_argument(
     '-p',
     '--password',
@@ -135,11 +141,13 @@ password_options.add_argument(
     "password file is provided, we'll use the REPLICAT_PASSWORD environment variable.",
     type=_read_bytes,
 )
-common_options.set_defaults(password=_get_environb('REPLICAT_PASSWORD'))
+common_options_parser.set_defaults(password=_get_environb('REPLICAT_PASSWORD'))
 
 
 def make_main_parser(*parent_parsers):
     parser = argparse.ArgumentParser(add_help=True)
+    parser.add_argument('--version', action='version', version=__version__)
+
     subparsers = parser.add_subparsers(dest='action', required=True)
 
     init_parser = subparsers.add_parser('init', parents=parent_parsers)
@@ -218,12 +226,12 @@ def make_main_parser(*parent_parsers):
     return parser
 
 
-def parser_from_backend_class(cls, *, inherit_common=True):
+def parser_from_backend_class(cls, *, inherit_common=True, missing=None):
     """Create a parser instance that inherits arguments from the common parser
     and adds arguments based on the class constructor signature
     """
     if inherit_common:
-        parent_parsers = [common_options]
+        parent_parsers = [common_options_parser]
     else:
         parent_parsers = []
 
@@ -242,9 +250,9 @@ def parser_from_backend_class(cls, *, inherit_common=True):
         help_text = f'or the {environment_var} environment variable'
         if arg.default is not arg.empty:
             default = arg.default
-            help_text += f', or the constructor default {default!r}'
+            help_text += f', or the constructor default ({default})'
         else:
-            default = None
+            default = missing
 
         name = name.replace('_', '-')
         group.add_argument(
@@ -258,28 +266,29 @@ def parser_from_backend_class(cls, *, inherit_common=True):
     return parser
 
 
-def parse_unknown_args(args_list):
-    stack = list(args_list)
+def parse_cli_settings(args_list):
     mapping = {}
-    group = []
+    unknown = []
+    flag = None
 
-    while stack:
-        last = stack.pop()
-        if last.startswith('--') and group:
-            if len(group) > 1:
-                group.reverse()
-                value = group
-            else:
-                (value,) = group
-
-            mapping[last.lstrip('-').replace('-', '_')] = value
-            group = []
-        elif last.startswith('-'):
-            group = []
+    for arg in args_list:
+        if arg.startswith('--'):
+            if flag is not None:
+                unknown.append(flag)
+            flag = arg
         else:
-            group.append(guess_type(last))
+            if flag is not None:
+                key = flag.lstrip('-').replace('-', '_')
+                value = guess_type(arg)
+                mapping[key] = value
+                flag = None
+            else:
+                unknown.append(arg)
 
-    return mapping
+    if flag is not None:
+        unknown.append(flag)
+
+    return mapping, unknown
 
 
 def adapter_from_config(name, **kwargs):
@@ -367,15 +376,6 @@ def type_reverse(object):
         return base64.standard_b64decode(encoded)
 
 
-def safe_kwargs(func, args):
-    params = inspect.signature(func).parameters
-    return {
-        name: args[name]
-        for name, arg in params.items()
-        if arg.kind is arg.KEYWORD_ONLY and name in args
-    }
-
-
 _async_auth_glock = asyncio.Lock()
 _async_auth_locks = weakref.WeakKeyDictionary()
 _sync_auth_glock = threading.Lock()
@@ -470,24 +470,6 @@ def disable_gc(func):
     return _decorator
 
 
-class async_client:
-
-    """Creates and stores an httpx.AsyncClient instance when accessed through
-    the parent object. It should only be accessed from within async methods.
-    """
-
-    def __init__(self, *args, **kwargs):
-        self.args, self.kwargs = args, kwargs
-
-    def __get__(self, instance, owner):
-        try:
-            return self._session
-        except AttributeError:
-            # loop should be picked up automatically
-            self._session = httpx.AsyncClient(*self.args, **self.kwargs)
-            return self._session
-
-
 class DefaultNamespace(SimpleNamespace):
     def __getattr__(self, attr):
         return None
@@ -579,3 +561,13 @@ def iter_chunks(file, chunk_size=128_000):
 async def aiter_chunks(file, chunk_size=128_000):
     async for chunk in async_gen_wrapper(iter_chunks(file, chunk_size=chunk_size)):
         yield chunk
+
+
+async def as_completed(tasks):
+    queue = asyncio.Queue()
+
+    for task in tasks:
+        task.add_done_callback(queue.put_nowait)
+
+    for _ in tasks:
+        yield await queue.get()
